use std::{cell::UnsafeCell, ops::Index, ptr::NonNull, sync::RwLock};

use rayon::prelude::ParallelIterator;

use crate::v2_types::{
    self,
    traits::{AsUObject, UObjectTrait},
};

lazy_static::lazy_static! {
    pub static ref GOBJECTS: RwLock<GObjects> = RwLock::new(GObjects::new());
}

#[derive(Debug, Default)]
pub struct GObjects {
    objects: Option<NonNull<UnsafeCell<ObjectArray>>>,
}

impl GObjects {
    pub const fn new() -> Self {
        Self { objects: None }
    }

    pub fn set_objects(&mut self, objects: NonNull<ObjectArray>) {
        self.objects = Some(objects.cast());
    }

    pub fn as_objects(&self) -> Option<&ObjectArray> {
        self.objects
            .map(|objects| unsafe { &*objects.as_ref().get() })
    }
}

unsafe impl Send for GObjects {}
unsafe impl Sync for GObjects {}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArrayItem {
    object: Option<v2_types::UObject>,
    sn: u32,
}

unsafe impl Send for ObjectArrayItem {}
unsafe impl Sync for ObjectArrayItem {}

impl ObjectArrayItem {
    pub fn object(&self) -> Option<v2_types::UObject> {
        self.object
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArrayInner {
    objects: NonNull<Option<NonNull<ObjectArrayItem>>>,
    pre_allocated_objects: Option<NonNull<ObjectArrayItem>>,
    max_elements: u32,
    num_elements: u32,
    max_chunks: u32,
    num_chunks: u32,
}

impl ObjectArrayInner {
    const ELEMENTS_PER_CHUNK: usize = 64 * 1024;

    fn chunks_as_slice(&self) -> &[Option<NonNull<ObjectArrayItem>>] {
        unsafe { std::slice::from_raw_parts(self.objects.as_ptr(), self.num_chunks as usize) }
    }

    pub fn get_chunks(&self) -> Vec<&[ObjectArrayItem]> {
        (0..self.num_chunks as usize)
            .filter_map(|i| self.chunk_as_slice(i))
            .collect()
    }

    fn get_chunk_size(&self, chunk_index: usize) -> Option<usize> {
        if chunk_index < self.num_chunks as usize {
            if chunk_index * Self::ELEMENTS_PER_CHUNK <= self.num_elements as usize {
                Some(Self::ELEMENTS_PER_CHUNK)
            } else {
                Some(self.num_elements as usize % Self::ELEMENTS_PER_CHUNK)
            }
        } else {
            None
        }
    }

    fn chunk_as_slice(&self, chunk_index: usize) -> Option<&[ObjectArrayItem]> {
        let chunks = self.chunks_as_slice();
        let chunk = unsafe {
            chunks[chunk_index].map(|ptr| {
                std::slice::from_raw_parts(ptr.as_ptr(), self.get_chunk_size(chunk_index).unwrap())
            })
        };

        chunk
    }

    /// returns the chunk index and the index inside that chunk, or none if the index is out of range
    fn get_chunked_index(&self, i: usize) -> Option<(usize, usize)> {
        if i < self.num_elements as usize {
            let chunk_index = i / Self::ELEMENTS_PER_CHUNK;
            let within_chunk_index = i % Self::ELEMENTS_PER_CHUNK;

            Some((chunk_index, within_chunk_index))
        } else {
            None
        }
    }

    pub fn get_index(&self, i: usize) -> Option<&ObjectArrayItem> {
        if let Some((chunk, within_chunk)) = self.get_chunked_index(i) {
            self.chunk_as_slice(chunk)?.get(within_chunk)
        } else {
            None
        }
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct ObjectArray {
    first_gc_index: u32,
    last_non_gc_index: u32,
    max_objects_not_considered_by_gc: u32,
    open_for_disregard_for_gc: bool,
    inner: ObjectArrayInner,
}

unsafe impl Send for ObjectArray {}
unsafe impl Sync for ObjectArray {}

impl ObjectArray {
    pub fn len(&self) -> usize {
        self.inner.num_elements as usize
    }

    pub fn iter(&self) -> iter::ObjectArrayIterator<'_> {
        iter::ObjectArrayIterator::new(self)
    }

    pub fn iter_package_objects(&self) -> impl Iterator<Item = &'_ ObjectArrayItem> {
        self.iter()
            .filter(|entry| matches!(entry.object(), Some(object) if object.is_package_object()))
    }

    pub fn par_iter(&self) -> par_iter::ObjectArrayChunkedParallelIterator<'_> {
        par_iter::ObjectArrayChunkedParallelIterator::new(self.inner.get_chunks())
    }

    pub fn par_iter_package_objects(&self) -> impl ParallelIterator<Item = &'_ ObjectArrayItem> {
        self.par_iter()
            .filter(|entry| matches!(entry.object(), Some(object) if object.is_package_object()))
    }
}

impl Index<usize> for ObjectArray {
    type Output = ObjectArrayItem;

    fn index(&self, i: usize) -> &Self::Output {
        self.inner.get_index(i).expect("Out of bounds access")
    }
}

pub mod iter {
    use super::{ObjectArray, ObjectArrayItem};

    pub struct ObjectArrayIterator<'a> {
        objects: &'a ObjectArray,
        index: usize,
    }

    impl<'a> ObjectArrayIterator<'a> {
        pub(super) fn new(objects: &'a ObjectArray) -> Self {
            Self { objects, index: 0 }
        }
    }

    impl<'a> Iterator for ObjectArrayIterator<'a> {
        type Item = &'a ObjectArrayItem;

        fn next(&mut self) -> Option<Self::Item> {
            let item = if self.index < self.objects.len() {
                Some(&self.objects[self.index])
            } else {
                None
            };

            self.index += 1;

            item
        }
    }

    impl<'a> IntoIterator for &'a ObjectArray {
        type Item = &'a ObjectArrayItem;
        type IntoIter = ObjectArrayIterator<'a>;

        fn into_iter(self) -> Self::IntoIter {
            ObjectArrayIterator {
                objects: self,
                index: 0,
            }
        }
    }
}

pub mod par_iter {
    pub use rayon::iter::{IntoParallelIterator, ParallelIterator};

    use super::ObjectArrayItem;

    pub struct ObjectArrayChunkedParallelIterator<'a> {
        chunks: Vec<&'a [ObjectArrayItem]>,
    }

    impl<'a> ObjectArrayChunkedParallelIterator<'a> {
        pub fn new(chunks: Vec<&'a [ObjectArrayItem]>) -> Self {
            Self { chunks }
        }
    }

    impl<'a> ParallelIterator for ObjectArrayChunkedParallelIterator<'a> {
        type Item = &'a ObjectArrayItem;

        fn drive_unindexed<C>(self, consumer: C) -> C::Result
        where
            C: rayon::iter::plumbing::UnindexedConsumer<Self::Item>,
        {
            self.chunks
                .into_par_iter()
                .flatten_iter()
                .drive_unindexed(consumer)
        }
    }
}

pub trait FindClass {
    fn find_class<S>(&self, class_name: S) -> Option<v2_types::UClass>
    where
        S: Into<String>;
}

impl FindClass for ObjectArray {
    fn find_class<S>(&self, class_name: S) -> Option<v2_types::UClass>
    where
        S: Into<String>,
    {
        use crate::v2_types::traits::UObjectNonConst;
        let class_name = class_name.into();

        self.iter()
            .filter_map(|item| item.object())
            .filter(|object| {
                object
                    .get_full_name()
                    .map(|name| name == class_name)
                    .unwrap_or(false)
            })
            .map(|object| (unsafe { object.cast() }))
            .next()
    }
}
